#include <gmp.h>

struct quotient_node
{
  struct quotient_node *next;
  mpz_t q;
};

struct quotient_stack
{
  struct quotient_node *top;
  struct quotient_node *bottom;
};

void
quotient_stack_init(struct quotient_stack *stack);

void
quotient_stack_clear(struct quotient_stack *stack);

void
quotient_stack_push(struct quotient_stack *stack,
		    const mpz_t q);

void
quotient_stack_pop(struct quotient_stack *stack,
		   mpz_t q);

/* Builds R by putting A on top of B */
void
quotient_stack_nappend(struct quotient_stack *R,
		       struct quotient_stack *A, struct quotient_stack *B);

struct jhgcd
{
  mpz_t a;
  mpz_t b;
#if 0
  mpz_t r;
#endif
  struct quotient_stack quotients;

  int sign;
  
  mpz_t R[4];
};

void
jhgcd_init(struct jhgcd *hgcd);

void
jhgcd_clear(struct jhgcd *hgcd);

void
jhgcd_check(struct jhgcd *hgcd);

void
jhgcd_check_more(struct jhgcd *hgcd,
		const mpz_t a, const mpz_t b);

int
jhgcd_slow(struct jhgcd *hgcd);

int
jhgcd(struct jhgcd *hgcd);

/* Quotient stack. If a is n bits, we get at most 2n quotients. The
   total size of the quotients, when we run-length encode runs of
   ones, size limited to about n, the number of bits of a.

   We let negative entries in the size array represent *runs* of q =
   1. Through out the code we represent q = 1 with qsize = 0. */

struct mpn_quotient_stack
{
  /* Points to the limb storage, size about size of a + size of b */
  mp_ptr limbs;
  mp_size_t alloc;

  mp_size_t *sizes;
  mp_size_t nsizes;

  /* Index of the current size entry */
  mp_size_t *next_size;

  /* When non-NULL, points to the current run count. */
  mp_size_t *run;
  
  /* The *end* of the current quotient, and the start of the next
     one. */
  mp_ptr top;
};

#define QUOTIENT_ROOM(s, qsize) \
((s)->top - (s)->limbs + (qsize) <= (s)->alloc)

mp_size_t
mpn_quotient_stack_limb_need(mp_size_t asize, mp_size_t bsize);

mp_size_t
mpn_quotient_stack_size_need(mp_size_t asize, mp_size_t bsize);

void
mpn_quotient_stack_init(struct mpn_quotient_stack *stack,
			mp_limb_t *limbs, mp_size_t alloc,
			mp_size_t *sizes, mp_size_t nsizes);

void
mpn_quotient_stack_push(struct mpn_quotient_stack *stack,
			mp_size_t qsize);

/* Quotient located at the top of the stack */
mp_size_t
mpn_quotient_stack_pop(struct mpn_quotient_stack *stack);

int
mpn_quotient_stack_equal(const struct mpn_quotient_stack *a,
			 const struct mpn_quotient_stack *b);

void
trace_mpn_quotients(const char *label,
		    const struct mpn_quotient_stack *stack);

struct mpn_jhgcd
{
  mp_ptr  ap; mp_size_t  asize;
  mp_ptr  bp; mp_size_t  bsize;
  mp_ptr r0p; mp_size_t r0size;
  mp_ptr r1p; mp_size_t r1size;
  
  /* Allocated space for the below values, for sanity checking. */
  mp_size_t alloc;

  /* If sign = 0, the diagonal elements u,v' are >=0, off-diagonal
     entries v, u' <= 0. If sign < 0, then diagonal <= 0, off-diagonal
     >= 0. */
  /* FIXME: sign is redundantm we should have sign < 0 iff current is
     odd. */
  int sign;
  mp_size_t uvsize[6]; mp_ptr uvp[6];
};

mp_size_t
mpn_jhgcd_itch(mp_size_t asize, mp_size_t bsize);

void
mpn_jhgcd_init(struct mpn_jhgcd *hgcd,
	       mp_srcptr ap, mp_size_t asize,
	       mp_srcptr bp, mp_size_t bsize,
	       mp_ptr tp, mp_size_t talloc);

int
mpn_jhgcd_equal(const struct mpn_jhgcd *R,
		const struct mpn_jhgcd *S);

void
trace_mpn_matrix(const char *label, const struct mpn_jhgcd *hgcd);

mp_size_t
mpn_gcd_jebelean(mp_ptr rp,
		 mp_ptr ap, mp_size_t asize,
		 mp_ptr bp, mp_size_t bsize,
		 mp_ptr tp);

mp_size_t
mpn_jhgcd_slow_itch(struct mpn_jhgcd *hgcd);

int
mpn_jhgcd_slow(struct mpn_jhgcd *hgcd,
	       struct mpn_quotient_stack *quotients,
	       mp_ptr tp, mp_size_t talloc);

mp_size_t
mpn_jhgcd_lehmer_itch(struct mpn_jhgcd *hgcd);

int
mpn_jhgcd_lehmer(struct mpn_jhgcd *hgcd,
		 struct mpn_quotient_stack *quotients,
		 mp_ptr tp, mp_size_t talloc);

mp_size_t
mpn_jhgcd_schoenhage_itch(struct mpn_jhgcd *hgcd);

int
mpn_jhgcd_schoenhage(struct mpn_jhgcd *hgcd,
		     struct mpn_quotient_stack *quotients,
		     mp_ptr tp, mp_size_t talloc);

mp_size_t
mpn_fgcd(mp_ptr rp,
	 mp_srcptr ap, mp_size_t asize,
	 mp_srcptr bp, mp_size_t bsize);

#if WANT_CHECKS
int
mpn_jhgcd_schoenhage_wrapper(struct mpn_jhgcd *hgcd,
			     struct mpn_quotient_stack *quotients,
			     mp_ptr tp, mp_size_t talloc);

#define mpn_jhgcd_schoenhage mpn_jhgcd_schoenhage_wrapper
#endif
